//*******************************************************************
//
//      BinaryTree.java			Authors:  Lewis/Chase
//
//     modified by Ed Hong, 11/9/04, to use ArrayList.
//      and to not use packages.
//	   Implements the BinaryTreeADT interface
//*******************************************************************

// package jss2; 
import java.util.Iterator;
import java.util.ArrayList;
// import jss2.exceptions.*; 

public class BinaryTree implements BinaryTreeADT 
{

   protected int count;
   protected BinaryTreeNode root; 

   //================================================================
   //  Creates an empty binary tree.
   //================================================================
   public BinaryTree() 
   {
      count = 0;
      root = null;
   }  // constructor BinaryTree

   //================================================================
   //  Creates a binary tree with the specified element as its root.
   //================================================================
   public BinaryTree (Object element) 
   {
      count = 1;
      root = new BinaryTreeNode (element);
   }  // constructor BinaryTree

   //================================================================
   //  Constructs a binary tree from the two specified binary trees.
   //================================================================
   public BinaryTree (Object element, BinaryTree leftSubtree,
                                      BinaryTree rightSubtree) 
   {

	 root = new BinaryTreeNode (element);
	 count = 1;
	 if (leftSubtree != null)
	 {
		count = count + leftSubtree.size();
		root.left = leftSubtree.root;
	 }//if
	 else
		root.left = null;
	 if (rightSubtree !=null)
	 {
		count = count + rightSubtree.size(); 
	 	root.right = rightSubtree.root;
	 }//if
	 else
		root.right = null;
 
   }  // constructor BinaryTree
   
   //================================================================
   //  Removes the left subtree of this binary tree.
   //================================================================
   public void removeLeftSubtree() 
   {
      if (root.left != null)
		count = count - root.left.numChildren() - 1;
      root.left = null;
   }  // method removeLeftSubtree

   //================================================================
   //  Removes the right subtree of this binary tree.
   //================================================================
   public void removeRightSubtree() 
   {
      if (root.right != null)
		count = count - root.right.numChildren() - 1; 
      root.right = null;
   }  // method removeRightSubtree
   
   //================================================================
   //  Deletes all nodes from the binary tree.
   //================================================================
   public void removeAllElements() 
   {
      count = 0;
      root = null;
   }  // method removeAllElements
   
   //================================================================
   //  Returns true if the binary tree is empty and false otherwise.
   //================================================================
   public boolean isEmpty() 
   {
      return (count == 0);
   }  // method isEmpty

   //================================================================
   //  Returns true if the binary tree is empty and false otherwise.
   //================================================================
   public int size() 
   {
      return count;
   }  // method size
   
   //================================================================
   //  Returns true if the tree contains an element that matches the
   //  specified target element and false otherwise.
   //================================================================
   public boolean contains (Object targetElement) 
   {

      Object temp;
      boolean found = false;

      try 
	 {
         temp = find (targetElement);
         found = true;
      }//try
  
      catch (Exception ElementNotFoundException) 
	 {
         found = false;
      }//catch

      return found;

   }  // method contains

   //================================================================
   //  Returns a reference to the specified target element if it is
   //  found in the binary tree.  Throws a NoSuchElementException if
   //  the specified target element is not found in the binary tree.
   //================================================================
   public Object find(Object targetElement) throws ElementNotFoundException {
     BinaryTreeNode current = findagain( targetElement, root );
     if( current == null )
       throw new ElementNotFoundException("binarytree");
     return current.element;
   } // method find

   //================================================================
   //  Returns a reference to the specified target element if it is
   //  found in the binary tree.
   //================================================================
   private BinaryTreeNode findagain(Object targetElement, BinaryTreeNode next) {
     if (next == null) {
       return null;
     }
     if (next.element.equals(targetElement)) {
       return next;
     }
     BinaryTreeNode temp = findagain(targetElement, next.left);
     if (temp == null) {
       temp = findagain(targetElement, next.right);
     }
     return temp;
   } // method findagain
 

   //================================================================
   //  Returns a string representation of the binary tree.
   //================================================================
   public String toString() 
   {
      ArrayList templist = new ArrayList();
      preorder (root, templist);
      return templist.toString();
   }  // method toString

   //================================================================
   //  Performs an inorder traversal on the binary tree by calling an
   //  overloaded, recursive inorder method that starts with
   //  the root.
   //================================================================
   public Iterator iteratorInOrder() 
   { 
      ArrayList templist = new ArrayList();
      inorder (root, templist);
      return templist.iterator();
   }  // method inorder

   //================================================================
   //  Performs a recursive inorder traversal, adding items to 
   //   templist as they are visited.
   //================================================================
   protected void inorder (BinaryTreeNode node, ArrayList templist) 
   { 
      if (node != null) 
	 {
         inorder (node.left, templist);
         templist.add(node.element); // adds to end of list.
         inorder (node.right, templist);
      }//if
       
   }  // method inorder

   //================================================================
   //  Performs an preorder traversal on the binary tree by calling an
   //  overloaded, recursive preorder method that starts with
   //  the root.
   //================================================================
   public Iterator iteratorPreOrder() 
   {
      ArrayList templist = new ArrayList();
      preorder (root, templist);
      return templist.iterator();
   }  // method preorder

   //================================================================
   //  Performs a recursive preorder traversal, adding items to 
   //    templist as they are visited
   //================================================================
   protected void preorder (BinaryTreeNode node, ArrayList templist) 
   {
      if (node != null) 
	 {
         templist.add(node.element); // adds to end of list.
         preorder (node.left, templist);
         preorder (node.right, templist);
      }//if
   }  // method preorder

   //================================================================
   //  Performs an postorder traversal on the binary tree by calling
   //  an overloaded, recursive postorder method that starts
   //  with the root.
   //================================================================
   public Iterator iteratorPostOrder() 
   {
      ArrayList templist = new ArrayList();
      postorder (root, templist);
      return templist.iterator();
   }  // method postorder

   //================================================================
   //  Performs a recursive postorder traversal.
   //================================================================
   protected void postorder (BinaryTreeNode node, ArrayList templist) 
   {
       if (node != null) 
        {
          postorder (node.left, templist);
          postorder (node.right, templist);
          templist.add(node.element);
       }//if
   }  // method postorder

   //================================================================
   //  Performs a levelorder traversal on the binary tree, using a
   //  templist.
   //================================================================
   public Iterator iteratorLevelOrder() 
   {

      ArrayList nodes = new ArrayList();
      ArrayList templist = new ArrayList();
      BinaryTreeNode current;

      nodes.add (root); // adds to end.

      while (! nodes.isEmpty()) 
	 {
         current = (BinaryTreeNode)nodes.remove(0); // removeFirst

         if (current != null) 
	    {
            templist.add(current.element);
            nodes.add (current.left);
            nodes.add (current.right);
         }//if
         else
            templist.add("null");
      }//while
      return templist.iterator();
   }  // method levelorder
}  // class BinaryTree

